//===--- ModularizeHeaderFile.h - Stores PP directives ----------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
///
/// \file
/// \brief Defines a class to store information about preprocessor directives
///   encountered in a header file which might be relevant to finding modules-related
///   problems.
///
//===----------------------------------------------------------------------===//

#ifndef MODULARIZEHEADERFILE_H
#define MODULARIZEHEADERFILE_H

#include "clang/Lex/Preprocessor.h"
#include "clang/Basic/SourceLocation.h"
#include "llvm/ADT/StringMap.h"
#include "ModularizePPDirective.h"
#include <string>

namespace Modularize {
  using namespace clang;
  using namespace llvm;

/// \brief This class stores information about preprocessor directives
///   encountered in a header file which might be relevant to finding
///   modules-related problems.
class ModularizeHeaderFile {
public:
  ModularizeHeaderFile(std::string name, Preprocessor &pp);
  ~ModularizeHeaderFile();

  // Name accessors.
  std::string& getName() { return Name; }
  void setName(std::string name) { Name = name; };

  // Preprocessor directive accessors.
  ModularizePPDirectiveVector &getDirectives() { return Directives; }
  int getDirectiveCount() { return Directives.size(); }
  ModularizePPDirectiveVectorIterator getDirectivesBegin()
    { return Directives.begin(); }
  ModularizePPDirectiveVectorIterator getDirectivesEnd()
    { return Directives.end(); }
  ModularizePPDirective &getDirective(int index) { return Directives[index]; }
  void addDirective(ModularizePPDirective &directive)
    { Directives.push_back(directive); };

  // Get highest line-numbered directive.
  unsigned getLastLineNumber();

  // Return true if this snapshot matches the other snapshot.
  bool match(ModularizeHeaderFile *other);

  // Display the information for debugging.
  void dump(int level = 0);
private:
  // The full path name of the header file.
  std::string Name;
  // The array of preprocessor directives in the file, limited
  // to the kinds modularize is interested in.
  ModularizePPDirectiveVector Directives;
};

// Collections for users of this typs.
typedef StringMap<ModularizeHeaderFile *> ModularizeHeaderFileMap;
typedef ModularizeHeaderFileMap::iterator ModularizeHeaderFileMapIterator;

}  // end namespace clang

#endif
