//==- ModularizeHeaderTracker.cpp - Represents header instances -*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//

#include "ModularizeHeaderTracker.h"
#include "ModularizeUtilities.h"
#include "llvm/Support/raw_ostream.h"

namespace Modularize {
using namespace llvm;
using namespace clang;

ModularizeHeaderTracker::ModularizeHeaderTracker(
      std::string& name, ModularizeHeaderFile *headerFile,
      std::string topHeader) :
    Name(name) {
  ModularizeHeaderInstance *headerInstance = new ModularizeHeaderInstance(
	headerFile, topHeader);
  addHeaderInstance(headerInstance);
}

ModularizeHeaderTracker::~ModularizeHeaderTracker() {
}

// Process a header snapshot.
// If the snapshot differs, add a new header instance.
void ModularizeHeaderTracker::addHeaderFile(
	ModularizeHeaderFile *headerFile, std::string topHeader) {
  ModularizeHeaderInstanceVectorIterator iter = HeaderInstances.begin();
  ModularizeHeaderInstanceVectorIterator EI = HeaderInstances.end();
  for (; iter != EI; ++iter) {
    ModularizeHeaderFile *existingHeader = (*iter)->getHeaderFile();
    if (existingHeader->match(headerFile)) {
		  (*iter)->addTopHeader(topHeader);
		  return;
	  }
  }
  ModularizeHeaderInstance *newInstance = new ModularizeHeaderInstance(
	headerFile, topHeader);
  addHeaderInstance(newInstance);
}

// Add a header instance.
void ModularizeHeaderTracker::addHeaderInstance(
	ModularizeHeaderInstance *headerInstance) {
  HeaderInstances.push_back(headerInstance);
}

// Report instances where header snapshots differ.
// Returns true if no errors, i.e. no differing snapshots.
bool ModularizeHeaderTracker::report() {
  int headerInstanceIndex;
  int headerInstanceCount = getHeaderInstanceCount();
  ModularizeHeaderInstance *headerInstance;
  ModularizeHeaderInstance *headerInstance0;
  ModularizeHeaderFile *headerFile;
  ModularizeHeaderFile *headerFile0;
  int directiveIndex;
  ModularizePPDirective *directive;
  ModularizePPDirective *directive0;
  bool mismatch;
  if (getHeaderInstanceCount() > 1) {
    int directiveCount = 0x7fffffff;  // Big number.
    ModularizeHeaderInstanceVectorIterator hiIter = getHeaderInstancesBegin();
    ModularizeHeaderInstanceVectorIterator hiEnd = getHeaderInstancesEnd();
    // Get the number of instances which is least.
	  for (; hiIter != hiEnd; ++hiIter) {
      headerInstance = (*hiIter);
      headerFile = headerInstance->getHeaderFile();
      if (headerFile->getDirectiveCount() < directiveCount)
        directiveCount = headerFile->getDirectiveCount();
	  }
    for (directiveIndex = 0; directiveIndex < directiveCount;
        directiveIndex++) {
      headerInstance0 = getHeaderInstance(0);
      headerFile0 = headerInstance0->getHeaderFile();
      directive0 = &headerFile0->getDirective(directiveIndex);
      mismatch = false;
      for (headerInstanceIndex = 1; headerInstanceIndex < headerInstanceCount;
          headerInstanceIndex++) {
        headerInstance = getHeaderInstance(headerInstanceIndex);
        headerFile = headerInstance->getHeaderFile();
        directive = &headerFile->getDirective(directiveIndex);
        if (!directive->match(directive0))
          mismatch = true;
      }
      if (mismatch) {
        errs() << "warning: The instances of header " << Name << " have different contents after preprocessing:\n";
        for (headerInstanceIndex = 0; headerInstanceIndex < headerInstanceCount;
            headerInstanceIndex++) {
          headerInstance = getHeaderInstance(headerInstanceIndex);
          headerFile = headerInstance->getHeaderFile();
          errs() << getIndent(1) << "When included or nested in these top-level headers:\n";
	        ModularizeTopHeaderVectorIterator thIter = headerInstance->getTopHeadersBegin();
	        ModularizeTopHeaderVectorIterator thEnd = headerInstance->getTopHeadersEnd();
          for (; thIter != thEnd; ++thIter) {
            errs() << getIndent(2) << *thIter << "\n";
          }
          errs() << getIndent(1) << "This preprocessor directive is mismatched (the parentheses show the condition after macro substitution):\n";
          directive = &headerFile->getDirective(directiveIndex);
          directive->dump(2, Name.c_str());
        }
        break;  // Stop here, because the blocks might be different after this.
      }
    }
    return false;   // There was a mismatch.
  }
  return true;      // No mismatch.
}

// Display the information for debugging.
void ModularizeHeaderTracker::dump(int level) {
  errs() << getIndent(level) << Name << ":\n";
  errs() << getIndent(level + 1) << "instance count: "
    << getHeaderInstanceCount() << "\n";

  errs() << getIndent(level + 1) << "instances: \n";
  ModularizeHeaderInstanceVectorIterator hiIter = getHeaderInstancesBegin();
  ModularizeHeaderInstanceVectorIterator hiEnd = getHeaderInstancesEnd();
	int hiIndex;
	for (hiIndex = 0; hiIter != hiEnd; ++hiIter, ++hiIndex) {
    ModularizeHeaderInstance *headerInstance = (*hiIter);
    errs() << getIndent(level + 2) << "instance: " << hiIndex << "\n";
    headerInstance->dump(level + 3);
	}
}

}

