//===--- PPCallbacks.cpp - Callbacks for Preprocessor actions ---*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//

#include "ModularizePPCallbacks.h"
#include "ModularizeUtilities.h"
#include "llvm/Support/raw_ostream.h"
#include "clang/Basic/SourceManager.h"
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>

namespace Modularize {
using namespace llvm;
using namespace clang;

ModularizePPCallbacks::ModularizePPCallbacks(Preprocessor& pp, StringRef topFile,
      ModularizeMasterHeaderTracker &masterHeaderTracker) :
    MasterHeaderTracker(masterHeaderTracker),
  TopFile(topFile.str()),
    PP(pp),
    RootHeaderFile(NULL),
    CurrentHeaderFile(NULL) {
  ModularizeHeaderFile *mhf = new ModularizeHeaderFile(topFile.str(), PP);
  addHeaderFile(mhf);
  CurrentHeaderFile = mhf;
  RootHeaderFile = mhf;
}

ModularizePPCallbacks::~ModularizePPCallbacks() {
}

void ModularizePPCallbacks::FileChanged(SourceLocation Loc, FileChangeReason Reason,
                         SrcMgr::CharacteristicKind FileType,
                         FileID PrevFID) {
  std::string locgetFileName(getSourceLocationFile(Loc));
  ModularizeHeaderFile *mhf = NULL;
  std::string *fileName = NULL;
  switch (Reason) {
    case EnterFile:
      fileName = &locgetFileName;
      break;
    case ExitFile:
      if (PrevFID.isInvalid())
        fileName = &TopFile;
      else
        fileName = &locgetFileName;
      break;
    case SystemHeaderPragma:
      return;
    case RenameFile:
      return;
    default:
      return;
  }
  if (fileName == NULL)
    return;
  mhf = getHeaderFile(*fileName);
  if (mhf == NULL) {
    mhf = new ModularizeHeaderFile(*fileName, PP);
    addHeaderFile(mhf);
    CurrentHeaderFile = mhf;
    if (!RootHeaderFile)
      RootHeaderFile = mhf;
  }
  else
    CurrentHeaderFile = mhf;
}

void ModularizePPCallbacks::EndOfMainFile() {
  // Add all the header snapshots to the master header file tracker.
  ModularizeHeaderFileMapIterator iter = HeaderFileMap.begin();
  ModularizeHeaderFileMapIterator end = HeaderFileMap.end();
  for (; iter != end; ++iter) {
    MasterHeaderTracker.addHeaderFile((*iter).second, TopFile);
  }
}

void ModularizePPCallbacks::MacroExpands(const Token &MacroNameTok, const MacroDirective *MD,
                          SourceRange Range, const MacroArgs *Args) {
  std::string macroName = MacroNameTok.getIdentifierInfo()->getName().str();
  std::string value;
  const MacroInfo *MI = PP.getMacroInfo(MacroNameTok.getIdentifierInfo());
  MacroInfo::tokens_iterator I = MI->tokens_begin();
  MacroInfo::tokens_iterator E = MI->tokens_end();
  for (; I != E; ++I)
    value += PP.getSpelling((*I));
  addSymbol(macroName, value);
}

void ModularizePPCallbacks::If(SourceLocation Loc, SourceRange ConditionRange) {
  std::string unconvertedConditional(getSourceSnippet(ConditionRange));
  std::string convertedConditional(getMacroSubstitutedString(unconvertedConditional));
  unsigned lineNumber = getSourceLocationLineNumber(Loc);
  if (CurrentHeaderFile->getLastLineNumber() < lineNumber) {
    ModularizePPDirective directive(
      MPPD_If, lineNumber,
      unconvertedConditional, convertedConditional);
    CurrentHeaderFile->addDirective(directive);
  }
}

void ModularizePPCallbacks::Elif(SourceLocation Loc, SourceRange ConditionRange,
                  SourceLocation IfLoc) {
  std::string unconvertedConditional(getSourceSnippet(ConditionRange));
  std::string convertedConditional(getMacroSubstitutedString(unconvertedConditional));
  unsigned lineNumber = getSourceLocationLineNumber(Loc);
  if (CurrentHeaderFile->getLastLineNumber() < lineNumber) {
    ModularizePPDirective directive(
      MPPD_ElIf, getSourceLocationLineNumber(Loc),
      unconvertedConditional, convertedConditional);
    CurrentHeaderFile->addDirective(directive);
  }
}

void ModularizePPCallbacks::Ifdef(SourceLocation Loc, const Token &MacroNameTok,
                   const MacroDirective *MD) {
  std::string macroName = MacroNameTok.getIdentifierInfo()->getName().str();
  std::string unconvertedConditional(macroName);
  std::string convertedConditional(
    macroName + (MD ? " (defined)" : " (not defined)"));
  unsigned lineNumber = getSourceLocationLineNumber(Loc);
  if (CurrentHeaderFile->getLastLineNumber() < lineNumber) {
    ModularizePPDirective directive(
      MPPD_IfDef, getSourceLocationLineNumber(Loc),
      unconvertedConditional, convertedConditional);
    CurrentHeaderFile->addDirective(directive);
  }
}

void ModularizePPCallbacks::Ifndef(SourceLocation Loc, const Token &MacroNameTok,
                    const MacroDirective *MD) {
  std::string macroName = MacroNameTok.getIdentifierInfo()->getName().str();
  std::string unconvertedConditional(macroName);
  std::string convertedConditional(
    macroName + (MD ? " (defined)" : " (not defined)"));
  unsigned lineNumber = getSourceLocationLineNumber(Loc);
  if (CurrentHeaderFile->getLastLineNumber() < lineNumber) {
    ModularizePPDirective directive(
      MPPD_IfNDef, getSourceLocationLineNumber(Loc),
      unconvertedConditional, convertedConditional);
    CurrentHeaderFile->addDirective(directive);
  }
}

// Some utility functions.

ModularizeHeaderFile *ModularizePPCallbacks::getHeaderFile(const std::string& name) {
  ModularizeHeaderFileMap::iterator iter = HeaderFileMap.find(name);
  if (iter != HeaderFileMap.end())
    return iter->second;
  return NULL;
}

void ModularizePPCallbacks::addHeaderFile(ModularizeHeaderFile *headerFile) {
  HeaderFileMap[headerFile->getName()] = headerFile;
}

// Add a symbol definition.  If the symbol already exists,
// its value is replaced.
void ModularizePPCallbacks::addSymbol(std::string name, std::string value) {
  Symbols[name] = value;
}

// Get symbol value.  Return empty string if not defined.
std::string ModularizePPCallbacks::getSymbol(std::string name) {
  ModularizeSymbolMapIterator iter = Symbols.find(name);
  if (iter != Symbols.end())
    return (*iter).second;
  return std::string();
}

// Get symbol value.  Return true if found.
bool ModularizePPCallbacks::getSymbol(std::string name, std::string &value) {
  ModularizeSymbolMapIterator iter = Symbols.find(name);
  if (iter != Symbols.end()) {
    value = (*iter).second;
  return true;
  }
  return false;
}

std::string ModularizePPCallbacks::getSourceLocationString(SourceLocation Loc) {
  if (Loc.isInvalid())
    return std::string("(none)");
  else
    return Loc.printToString(PP.getSourceManager());
}

std::string ModularizePPCallbacks::getSourceLocationFile(SourceLocation Loc) {
  std::string source(getSourceLocationString(Loc));
  int offset = source.find(':', 2);
  if (offset < 0)
    return source;
  return source.substr(0, offset);
}

unsigned ModularizePPCallbacks::getSourceLocationLineNumber(SourceLocation Loc) {
  return PP.getSourceManager().getPresumedLineNumber(Loc);
}

// Retrieve source snippet from file image.
std::string ModularizePPCallbacks::getSourceSnippet(SourceRange sourceRange) {
  SourceLocation bLoc(sourceRange.getBegin());
  SourceLocation eLoc(sourceRange.getEnd());

  // Decompose the locations into FID/Offset pairs.
  std::pair<FileID, unsigned> bLocInfo = PP.getSourceManager().getDecomposedLoc(bLoc);
  std::pair<FileID, unsigned> eLocInfo = PP.getSourceManager().getDecomposedLoc(eLoc);
  FileID FID = bLocInfo.first;
  unsigned bFileOffset = bLocInfo.second;
  unsigned eFileOffset = eLocInfo.second;
  unsigned length = eFileOffset - bFileOffset;

  // Get information about the buffer it points into.
  bool Invalid = false;
  const char *BufStart = PP.getSourceManager().getBufferData(FID, &Invalid).data();
  if (Invalid)
    return std::string();

  // Rewind from the current position to the start of the line.
  const char *bPtr = BufStart + bFileOffset;

  // Trim snippet.
  while ((*bPtr <= ' ') && (length != 0)) {
    bPtr++;
    length--;
  }

  while ((length != 0) && (bPtr[length - 1] <= ' '))
    length--;

  std::string returnValue(bPtr, length);

  return returnValue;
}

// Do macro substitutions in a string.
// FIXME: Note that this is very simple parser, such that it doesn't
// support things like function-style macros, macros with string
// or character literals, and defined().
// FIXME: There probably should be a function in the Preprocessor
// class that will lex a string, do substitutions, and return either
// a token list or a converted string.  However, it's somewhat
// problematic in that it could interfere with the preprocessor
// and lexor state, and call the MacroExpanded handlers again.
// So for now, we stick with this simple function, to be extend
// to support function-style macros.
std::string ModularizePPCallbacks::getMacroSubstitutedString(std::string str) {
  enum LexMode {
    LM_Top,
    LM_Symbol
  } lexMode = LM_Top;
  char output[1024];
  size_t outputLength = 0;
  char token[256];
  size_t tokenLength = 0;
  std::string symbolValue;
  size_t symbolLength;
  size_t length = str.length();
  size_t index;
  char c = '\0';
  for (index = 0; index <= length; index++) {
    if (index < length)
      c = str[index];
    else
      c = '\0';
restart:
    switch (lexMode) {
      case LM_Top:
        tokenLength = 0;
        if (isalpha(c) || (c == '_')) {
          lexMode = LM_Symbol;
          goto restart;
        }
        else {
          if (outputLength < sizeof(output) + 1)
            output[outputLength++] = c;
        }
        break;
      case LM_Symbol:
        if (isalpha(c) || (c == '_') || isdigit(c)) {
          if (tokenLength < sizeof(token) - 1)
            token[tokenLength++] = c;
        }
        else {
          token[tokenLength] = '\0';
          if (getSymbol(token, symbolValue)) {
            // Recursively do substitutions on macro value.
            symbolValue = getMacroSubstitutedString(symbolValue);
            symbolLength = symbolValue.length();
            if (outputLength + symbolLength < sizeof(output) - 1) {
              strcpy(output + outputLength, symbolValue.c_str());
              outputLength += symbolLength;
            }
            else {
              strncpy(output + outputLength, symbolValue.c_str(), sizeof(output) - outputLength - 1);
              outputLength = sizeof(output) - 1;
              output[sizeof(output) - 1] = '\0';
            }
          }
          else {
            if (outputLength + tokenLength < sizeof(output) - 1) {
              strcpy(output + outputLength, token);
              outputLength += tokenLength;
            }
            else {
              strncpy(output + outputLength, token, sizeof(output) - outputLength - 1);
              outputLength = sizeof(output) - 1;
              output[sizeof(output) - 1] = '\0';
            }
          }
          lexMode = LM_Top;
          goto restart;
        }
        break;
    }
  }
  output[outputLength] = '\0';
  return std::string(output);
}

}
