//===--- ModularizePPDirective.cpp - Represents a PP directive --*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//

#include "ModularizePPDirective.h"
#include "ModularizeUtilities.h"
#include "clang/Basic/SourceManager.h"
#include "llvm/Support/raw_ostream.h"

namespace Modularize {
using namespace llvm;
using namespace clang;

ModularizePPDirective::ModularizePPDirective(
	ModularizePPDirectiveKind directiveKind, unsigned lineNumber,
    std::string unconvertedConditional, std::string convertedConditional) :
  DirectiveKind(directiveKind),
  LineNumber(lineNumber),
  UnconvertedConditional(unconvertedConditional),
  ConvertedConditional(convertedConditional) {
}

ModularizePPDirective::~ModularizePPDirective() {
}

// Return true if this directive matches another.
// (It compares the converted conditionals.)
bool ModularizePPDirective::match(ModularizePPDirective *other) {
  if (ConvertedConditional == other->getConvertedConditional()) {
    return true;
  }
  return false;
}

// Print this directive to a stream, suitable for a message.
void ModularizePPDirective::print(raw_ostream &OS,
    int level, std::string fileName) {
  const char *directiveSpelling = getDirectiveSpelling(DirectiveKind);
  OS << getIndent(level) << fileName << ":" << LineNumber << ":1: #"
    << directiveSpelling << " " << UnconvertedConditional << "  (#" <<
    directiveSpelling << " " << ConvertedConditional << ")\n";
}

// Print this directive to a string, suitable for a message.
std::string ModularizePPDirective::printToString(
    int level, std::string fileName) {
  std::string S;
  llvm::raw_string_ostream OS(S);
  print(OS, level, fileName);
  return OS.str();
}

// Display useful information about this object for debugging.
void ModularizePPDirective::dump(int level, const char *fileName) {
  if (fileName == NULL)
    fileName = "line";
  print(errs(), level, fileName);
}

// Get directive spelling.
const char *ModularizePPDirective::getDirectiveSpelling(
    ModularizePPDirectiveKind kind) {
  const char *directiveString;
  switch (kind) {
    case MPPD_If:
      directiveString = "if";
      break;
    case MPPD_ElIf:
      directiveString = "elif";
      break;
    case MPPD_IfDef:
      directiveString = "ifdef";
      break;
    case MPPD_IfNDef:
      directiveString = "ifndef";
      break;
    default:
      directiveString = "(unknown)";
      break;
  }
  return directiveString;
}

}
