//===--- ModularizePPDirective.h - Represents a PP directive ----*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
///
/// \file
/// \brief Defines a class to store information about one preprocessor directive
///   that modularize might be interested in.
///
//===----------------------------------------------------------------------===//

#ifndef MODULARIZEPPDIRECTIVE_H
#define MODULARIZEPPDIRECTIVE_H

#include "clang/Lex/Preprocessor.h"
#include "clang/Basic/SourceLocation.h"
#include <string>

namespace Modularize {
  using namespace clang;
  using namespace llvm;

// Represents a preprocessor directive kind.
enum ModularizePPDirectiveKind {
  MPPD_If,
  MPPD_ElIf,
  MPPD_IfDef,
  MPPD_IfNDef
};

/// \brief This class stores information about one preprocessor directive
///   that modularize might be interested in.
class ModularizePPDirective {
public:
  ModularizePPDirective(
	  ModularizePPDirectiveKind directiveKind, unsigned lineNumber,
    std::string unconvertedConditional, std::string convertedConditional);
  ~ModularizePPDirective();

  // Source location string accessors.
  unsigned getLineNumber() { return LineNumber; }
  void setLineNumber(unsigned value) { LineNumber = value; };

  // Unconverted conditional accessors.
  std::string& getUnconvertedConditional() { return UnconvertedConditional; }
  void setUnconvertedConditional(std::string value) { UnconvertedConditional = value; };

  // Converted conditional accessors.
  std::string& getConvertedConditional() { return ConvertedConditional; }
  void setConvertedConditional(std::string value) { ConvertedConditional = value; };

  // Return true if this directive matches another.
  // (It compares the converted conditionals.)
  bool match(ModularizePPDirective *other);

  // Print this directive to a stream.
  void print(raw_ostream &OS, int level, std::string fileName);

  // Print this directive to a string suitable for a message.
  std::string printToString(int level, std::string fileName);

  // Display the information for debugging.
  void dump(int level = 0, const char *fileName = NULL);

  // Get directive spelling.
  static const char *getDirectiveSpelling(ModularizePPDirectiveKind kind);
private:
  // The kind of directive.
  ModularizePPDirectiveKind DirectiveKind;
  // Presumed line number for direcive.
  unsigned LineNumber;
  // Raw value representation of conditional.
  std::string UnconvertedConditional;
  // The conditional with macro substitutions done.
  std::string ConvertedConditional;
};

// Collections for users of this typs.
typedef std::vector<ModularizePPDirective> ModularizePPDirectiveVector;
typedef ModularizePPDirectiveVector::iterator ModularizePPDirectiveVectorIterator;

}  // end namespace clang

#endif
